#ifndef SORTED_LIST_H
#define SORTED_LIST_H

#include <OS.h>

/*
Usage:
	Handy for storing a large number of objects containing some form of
	unique id, FindItem() uses a binary search algorithm to allow speedy
	access to a specific object.
	Adding and removing items it O(n).
*/
/*
'Class' requirements:
	bool operator <  ( Class & ) const;
	bool operator <  ( Index & ) const;
	bool operator == ( Index & ) const;
*/
template <class Class, class Index>
class SortedList
{
	public:
		SortedList();
		~SortedList();
		
		void		AddItem( const Class * );
		void		RemoveItem( const Class * );
		Class *		ItemAt( int32 );
		
		int32		NumItems();
		Class *		FindItem( Index );
		int32		IndexOf( const Class * );
		
		void		RemoveAll();
		
	protected:
		int32		m_num_items;
		int32		m_buf_size;
		const Class **	m_buffer;
};

template <class Class, class Index>
SortedList<Class,Index>::SortedList()
:	m_num_items(0),
	m_buf_size(4)
{
	m_buffer = new const Class*[m_buf_size];
}

template <class Class, class Index>
SortedList<Class,Index>::~SortedList()
{
	delete [] m_buffer;
}

template <class Class, class Index>
void
SortedList<Class,Index>::AddItem( const Class * a )
{
	if ( m_num_items >= m_buf_size )
	{ // enlarge buffer
		m_buf_size *= 2;
		const Class ** temp = new const Class*[m_buf_size];
		memcpy( temp, m_buffer, (m_buf_size/2)*sizeof(Class*) );
		
		delete [] m_buffer;
		
		m_buffer = temp;
	}
	
	int32 insert_at = 0;
	
	for (; insert_at < m_num_items && *m_buffer[insert_at] < *a; insert_at++ )
		;
	
	memmove( 
		&m_buffer[insert_at+1], 
		&m_buffer[insert_at], 
		(m_num_items-insert_at)*sizeof(Class*)
	);
	
	m_buffer[insert_at] = a;
	m_num_items++;
}

template <class Class, class Index>
void
SortedList<Class,Index>::RemoveItem( const Class * a )
{
	int32 index = 0;
	for (; index < m_num_items && m_buffer[index] != a; index++)
		;
	
	if ( index < m_num_items )
	{ // target found, remove
		m_num_items--;
		
		memmove(
			&m_buffer[index],
			&m_buffer[index+1],
			(m_num_items-index)*sizeof(Class*)
		);
	}
}

template <class Class, class Index>
Class *
SortedList<Class,Index>::ItemAt( int32 a )
{
	return (Class*)m_buffer[a];
}

template <class Class, class Index>
int32
SortedList<Class,Index>::NumItems()
{
	return m_num_items;
}

template <class Class, class Index>
Class *
SortedList<Class,Index>::FindItem( Index index )
{
	if ( m_num_items == 0 )
		return NULL;
	
	// binary search
	int32 left = 0;
	int32 right = m_num_items-1;
	int32 center;
//	int32 steps = 0;
	
	while ( left != right )
	{
//		steps++;
		center = (left+right)/2;
		if ( (*m_buffer[center]) < index )
		{ // [l  t  c  t   r]
			left = center + 1;
		} else
		{
			right = center;
		}
//		printf("l: %ld, c: %ld, r: %ld \n", left, center, right);
	}

//	printf("SortedList::FindItem() completed in %ld steps\n", steps);

	if ( *m_buffer[left] == index )
		return (Class*)m_buffer[left];
	else
		return NULL;
}

template <class Class, class Index>
int32
SortedList<Class,Index>::IndexOf( const Class * a )
{
	int32 index = 0;
	for (; index < m_num_items && m_buffer[index] != a; index++)
		;
	
	if ( index < m_num_items )
	{ // target found, remove
		return index;
	} else
	{
		return -1;
	}
}

template <class Class, class Index>
void
SortedList<Class,Index>::RemoveAll()
{
	m_num_items = 0;
}

#endif
